import os
import re
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, DefaultDict, Dict, Iterator, List, Optional, Set
from uuid import UUID, uuid4

import tifffile

from panimg.exceptions import (
    UnconsumedFilesException,
    ValidationError,
)
from panimg.models import (
    ColorSpace,
    TIFFImage,
)

try:
    import openslide
except OSError:
    openslide = False

try:
    import pyvips
except OSError:
    pyvips = False


@dataclass
class GrandChallengeTiffFile:
    path: Path
    pk: UUID = field(default_factory=uuid4)
    image_width: int = 0
    image_height: int = 0
    resolution_levels: int = 0
    color_space: Optional[ColorSpace] = None
    voxel_width_mm: float = 0
    voxel_height_mm: float = 0
    associated_files: List[Path] = field(default_factory=list)

    def validate(self) -> None:
        if not self.image_width:
            raise ValidationError(
                "Not a valid tif: Image width could not be determined"
            )
        if not self.image_height:
            raise ValidationError(
                "Not a valid tif: Image height could not be determined"
            )
        if not self.resolution_levels:
            raise ValidationError(
                "Not a valid tif: Resolution levels not valid"
            )
        if not self.voxel_width_mm:
            raise ValidationError(
                "Not a valid tif: Voxel width could not be determined"
            )
        if not self.voxel_height_mm:
            raise ValidationError(
                "Not a valid tif: Voxel height could not be determined"
            )
        if not self.color_space:
            raise ValidationError(
                "Not a valid tif: Color space could not be determined"
            )


def _get_tag_value(tags, tag):
    try:
        return tags[tag].value
    except (KeyError, AttributeError):
        return None


def _get_voxel_spacing_mm(tags, tag):
    """
    Calculate the voxel spacing in mm.

    Use the set of tags from the tiff image to calculate the spacing for a
    particular dimension. Supports INCH and CENTIMETER resolution units.

    Parameters
    ----------
    tags
        The collection of tags from the tif file.
    tag
        The tag that contains the resolution of the tif file along the
        dimension of interest.

    Raises
    ------
    ValidationError
        Raised if an unrecognised resolution unit is used.


    Returns
    -------
        The voxel spacing in mm.

    """
    try:
        # Resolution is a tuple of the number of pixels and the length of the
        # image in cm or inches, depending on resolution unit
        resolution_unit = str(_get_tag_value(tags, "ResolutionUnit"))
        resolution = _get_tag_value(tags, tag)
        if resolution_unit == "RESUNIT.INCH":
            return 25.4 / (resolution[0] / resolution[1])
        elif resolution_unit == "RESUNIT.CENTIMETER":
            return 10 / (resolution[0] / resolution[1])
        raise ValidationError(
            f"Invalid resolution unit {resolution_unit}" f" in tiff file"
        )
    except (ZeroDivisionError, TypeError, IndexError) as e:
        raise ValidationError("Invalid resolution in tiff file") from e


def _extract_openslide_properties(
    *, gc_file: GrandChallengeTiffFile, image
) -> GrandChallengeTiffFile:
    if not gc_file.voxel_width_mm and "openslide.mpp-x" in image.properties:
        gc_file.voxel_width_mm = (
            float(image.properties["openslide.mpp-x"]) / 1000
        )
    if not gc_file.voxel_height_mm and "openslide.mpp-y" in image.properties:
        gc_file.voxel_height_mm = (
            float(image.properties["openslide.mpp-y"]) / 1000
        )
    if (
        not gc_file.image_height
        and "openslide.level[0].height" in image.properties
    ):
        gc_file.image_height = int(
            image.properties["openslide.level[0].height"]
        )

    if (
        not gc_file.image_width
        and "openslide.level[0].width" in image.properties
    ):
        gc_file.image_width = int(image.properties["openslide.level[0].width"])

    if (
        not gc_file.resolution_levels
        and "openslide.level-count" in image.properties
    ):
        gc_file.resolution_levels = int(
            image.properties["openslide.level-count"]
        )
    return gc_file


def _extract_tags(
    *, gc_file: GrandChallengeTiffFile, pages: tifffile.tifffile.TiffPages
) -> GrandChallengeTiffFile:
    """
    Extracts tags form a tiff file loaded with tifffile for use in grand challenge
    :param pages: The pages and tags from tifffile
    :return: The GrandChallengeTiffFile with the properties defined in the tags
    """
    if not pages:
        return gc_file

    tags = pages[0].tags

    gc_file.resolution_levels = len(pages)

    gc_file.color_space = _get_color_space(
        color_space_string=str(
            _get_tag_value(tags, "PhotometricInterpretation")
        )
    )
    gc_file.image_width = _get_tag_value(tags, "ImageWidth")
    gc_file.image_height = _get_tag_value(tags, "ImageLength")

    #  some formats like the Philips tiff don't have the spacing in their tags,
    #  we retrieve them later with OpenSlide
    if "XResolution" in tags:
        gc_file.voxel_width_mm = _get_voxel_spacing_mm(tags, "XResolution")
        gc_file.voxel_height_mm = _get_voxel_spacing_mm(tags, "YResolution")

    return gc_file


def _get_color_space(*, color_space_string) -> Optional[ColorSpace]:
    color_space_string = color_space_string.split(".")[1].upper()

    if color_space_string == "MINISBLACK":
        color_space = ColorSpace.GRAY
    else:
        try:
            color_space = ColorSpace[color_space_string]
        except KeyError:
            return None

    return color_space


def _load_with_tiff(
    *, gc_file: GrandChallengeTiffFile
) -> GrandChallengeTiffFile:
    tiff_file = tifffile.TiffFile(str(gc_file.path.absolute()))
    gc_file = _extract_tags(gc_file=gc_file, pages=tiff_file.pages)
    return gc_file


def _load_with_openslide(
    *, gc_file: GrandChallengeTiffFile
) -> GrandChallengeTiffFile:
    open_slide_file = openslide.open_slide(str(gc_file.path.absolute()))
    gc_file = _extract_openslide_properties(
        gc_file=gc_file, image=open_slide_file
    )
    return gc_file


mirax_pattern = r"INDEXFILE\s?=\s?.|FILE_\d+\s?=\s?."
# vms (and vmu) files contain key value pairs, where the ImageFile keys
# can have the following format:
# ImageFile =, ImageFile(x,y) ImageFile(z,x,y)
vms_pattern = (
    r"ImageFile(\(\d*,\d*(,\d)?\))?\s?=\s?."
    r"|MapFile\s?=\s?."
    r"|OptimisationFile\s?=\s?."
    r"|MacroImage\s?=\s?."
)


def _get_mrxs_files(mrxs_file: Path):
    # Find Slidedat.ini, which provides us with all the other file names
    name, _ = os.path.splitext(mrxs_file.name)
    slide_dat = mrxs_file.parent / name / "Slidedat.ini"
    file_matched = []
    if not slide_dat.exists():
        slide_dat = [
            f
            for f in slide_dat.parent.iterdir()
            if f.name.lower() == slide_dat.name.lower()
        ][0]
    with open(slide_dat) as f:
        lines = [
            line for line in f.readlines() if re.match(mirax_pattern, line)
        ]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(slide_dat.parent / original_name):
                file_matched.append(slide_dat.parent / original_name)
        file_matched.append(slide_dat)
    return file_matched


def _get_vms_files(vms_file: Path):
    file_matched = []
    with open(str(vms_file.absolute())) as f:
        lines = [line for line in f.readlines() if re.match(vms_pattern, line)]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(vms_file.parent / original_name):
                file_matched.append(vms_file.parent / original_name)
        return file_matched


def _convert(
    *,
    files: List[Path],
    associated_files_getter: Optional[Callable[[Path], List[Path]]],
    converter,
    output_directory: Path,
    file_errors: Dict[Path, List[str]],
) -> List[GrandChallengeTiffFile]:
    converted_files: List[GrandChallengeTiffFile] = []
    associated_files: List[Path] = []

    for file in files:
        try:
            gc_file = GrandChallengeTiffFile(file)

            if associated_files_getter:
                associated_files = associated_files_getter(gc_file.path)

            tiff_file = _convert_to_tiff(
                path=file,
                pk=gc_file.pk,
                converter=converter,
                output_directory=output_directory,
            )
        except Exception as e:
            file_errors[file].append(str(e))
            continue
        else:
            gc_file.path = tiff_file
            gc_file.associated_files = associated_files
            gc_file.associated_files.append(file)
            converted_files.append(gc_file)

    return converted_files


def _convert_to_tiff(
    *, path: Path, pk: UUID, converter, output_directory: Path
) -> Path:
    new_file_name = output_directory / path.name / f"{pk}.tif"
    new_file_name.parent.mkdir()

    image = converter.Image.new_from_file(
        str(path.absolute()), access="sequential"
    )

    using_old_vips = (
        pyvips.base.version(0) == 8 and pyvips.base.version(1) < 10
    )
    if (
        using_old_vips
        and image.get("xres") == 1
        and "openslide.mpp-x" in image.get_fields()
    ):
        # correct xres and yres if they have default value of 1
        # due to a bug that is resolved in pyvips 8.10
        x_res = 1000.0 / float(image.get("openslide.mpp-x"))
        y_res = 1000.0 / float(image.get("openslide.mpp-y"))
        image = image.copy(xres=x_res, yres=y_res)

    image.write_to_file(
        str(new_file_name.absolute()),
        tile=True,
        pyramid=True,
        bigtiff=True,
        compression="jpeg",
        Q=70,
    )

    return new_file_name


def _load_gc_files(
    *,
    files: Set[Path],
    converter,
    output_directory: Path,
    file_errors: Dict[Path, List[str]],
) -> List[GrandChallengeTiffFile]:
    loaded_files: List[GrandChallengeTiffFile] = []

    complex_file_handlers = {
        ".mrxs": _get_mrxs_files,
        ".vms": _get_vms_files,
        ".vmu": _get_vms_files,
        ".svs": None,
        ".ndpi": None,
        ".scn": None,
        ".bif": None,
    }

    for ext, handler in complex_file_handlers.items():
        complex_files = [file for file in files if file.suffix.lower() == ext]
        if len(complex_files) > 0:
            converted_files = _convert(
                files=complex_files,
                associated_files_getter=handler,
                converter=converter,
                output_directory=output_directory,
                file_errors=file_errors,
            )
            loaded_files += converted_files

    # don't handle files that are associated files
    for file in files:
        if not any(g.path == file for g in loaded_files) and not any(
            file in g.associated_files
            for g in loaded_files
            if g.associated_files is not None
        ):
            gc_file = GrandChallengeTiffFile(file)
            loaded_files.append(gc_file)

    return loaded_files


def image_builder_tiff(  # noqa: C901
    *, files: Set[Path]
) -> Iterator[TIFFImage]:
    if openslide is False:
        raise ImportError(
            f"Could not import openslide, which is required for the "
            f"{__name__} image builder. Either ensure that libopenslide-dev "
            f"is installed or remove {__name__} from your list of builders."
        )

    if pyvips is False:
        raise ImportError(
            f"Could not import pyvips, which is required for the "
            f"{__name__} image builder. Either ensure that libvips-dev "
            f"is installed or remove {__name__} from your list of builders."
        )

    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    with TemporaryDirectory() as output_directory:

        loaded_files = _load_gc_files(
            files=files,
            converter=pyvips,
            output_directory=Path(output_directory),
            file_errors=file_errors,
        )

        for gc_file in loaded_files:
            # try and load image with tiff file
            try:
                gc_file = _load_with_tiff(gc_file=gc_file)
            except Exception:
                file_errors[gc_file.path].append(
                    "Could not open file with tifffile."
                )

            # try and load image with open slide
            try:
                gc_file = _load_with_openslide(gc_file=gc_file)
            except Exception:
                file_errors[gc_file.path].append(
                    "Could not open file with OpenSlide."
                )

            # validate
            try:
                gc_file.validate()
                if gc_file.color_space is None:
                    # TODO This needs to be solved by refactoring of
                    # GrandChallengeTiffFile
                    raise RuntimeError("Color space not found")
            except ValidationError as e:
                file_errors[gc_file.path].append(f"Validation error: {e}.")
                continue

            if gc_file.associated_files:
                consumed_files = {
                    f.absolute() for f in gc_file.associated_files
                }
            else:
                consumed_files = {gc_file.path.absolute()}

            yield TIFFImage(
                file=gc_file.path,
                name=gc_file.path.name,
                consumed_files=consumed_files,
                width=gc_file.image_width,
                height=gc_file.image_height,
                voxel_width_mm=gc_file.voxel_width_mm,
                voxel_height_mm=gc_file.voxel_height_mm,
                resolution_levels=gc_file.resolution_levels,
                color_space=gc_file.color_space,
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



"""
Image builder for MetaIO mhd/mha files.

See: https://itk.org/Wiki/MetaIO/Documentation
"""
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, Iterator, List, Mapping, Set, Union

from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.image_builders.metaio_utils import (
    load_sitk_image,
    parse_mh_header,
)
from panimg.models import SimpleITKImage


def image_builder_mhd(  # noqa: C901
    *, files: Set[Path]
) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path: Path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A tuple of
     - all detected images
     - files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    element_data_file_key = "ElementDataFile"

    def detect_mhd_file(headers: Dict[str, str], path: Path) -> bool:
        try:
            data_file = headers[element_data_file_key]
        except KeyError:
            return False

        if data_file == "LOCAL":
            return False

        data_file_path = (path / Path(data_file)).resolve(strict=False)

        if path not in data_file_path.parents:
            raise ValueError(
                f"{element_data_file_key} references a file which is not in "
                f"the uploaded data folder"
            )
        if not data_file_path.is_file():
            raise ValueError("Data container of mhd file is missing")
        return True

    def detect_mha_file(headers: Mapping[str, Union[str, None]]) -> bool:
        data_file = headers.get(element_data_file_key, None)
        return data_file == "LOCAL"

    def format_error(message: str) -> str:
        return f"Mhd image builder: {message}"

    for file in files:
        try:
            parsed_headers = parse_mh_header(file)
        except ValidationError as e:
            file_errors[file].append(
                format_error(f"Could not validate or parse ITK header. {e}")
            )
            continue

        try:
            is_hd_or_mha = detect_mhd_file(
                parsed_headers, file.parent
            ) or detect_mha_file(parsed_headers)
        except ValueError as e:
            file_errors[file].append(format_error(str(e)))
            continue

        if is_hd_or_mha:
            file_dependency = None

            if parsed_headers[element_data_file_key] != "LOCAL":
                file_dependency = (
                    file.parent / parsed_headers[element_data_file_key]
                )
                if not file_dependency.is_file():
                    file_errors[file].append(
                        format_error("Cannot find data file")
                    )
                    continue

            try:
                simple_itk_image = load_sitk_image(file.absolute())
            except RuntimeError:
                file_errors[file].append(
                    format_error("SimpleITK could not open file.")
                )
                continue
            except ValidationError as e:
                file_errors[file].append(format_error(str(e)))
                continue

            consumed_files = {file}
            if file_dependency is not None:
                consumed_files.add(file_dependency)

            yield SimpleITKImage(
                image=simple_itk_image,
                name=file.name,
                consumed_files=consumed_files,
                spacing_valid=True,
            )
        else:
            file_errors[file].append(format_error("Not an ITK file"))
            continue

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



import re
import struct

import numpy as np
from construct import (
    PaddedString,
    Int16un,
    Struct,
    Int32sn,
    Int32un,
    Int8un,
)
from panimg.contrib.oct_converter.image_types import (
    OCTVolumeWithMetaData,
    FundusImageWithMetaData,
)
from pathlib import Path


class E2E:
    """ Class for extracting data from Heidelberg's .e2e file format.

        Notes:
            Mostly based on description of .e2e file format here:
            https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

        Attributes:
            filepath (str): Path to .img file for reading.
            header_structure (obj:Struct): Defines structure of volume's header.
            main_directory_structure (obj:Struct): Defines structure of volume's main directory.
            sub_directory_structure (obj:Struct): Defines structure of each sub directory in the volume.
            chunk_structure (obj:Struct): Defines structure of each data chunk.
            image_structure (obj:Struct): Defines structure of image header.
    """

    def __init__(self, filepath):
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

        self.chunk_structure = Struct(
            "magic" / PaddedString(12, "ascii"),
            "unknown" / Int32un,
            "unknown2" / Int32un,
            "pos" / Int32un,
            "size" / Int32un,
            "unknown3" / Int32un,
            "patient_id" / Int32un,
            "study_id" / Int32un,
            "series_id" / Int32un,
            "slice_id" / Int32sn,
            "ind" / Int16un,
            "unknown4" / Int16un,
            "type" / Int32un,
            "unknown5" / Int32un,
        )
        self.image_structure = Struct(
            "size" / Int32un,
            "type" / Int32un,
            "unknown" / Int32un,
            "width" / Int32un,
            "height" / Int32un,
        )
        self.lat_structure = Struct(
            "unknown" / PaddedString(14, "ascii"),
            "laterality" / Int8un,
            "unknown2" / Int8un,
        )

    def find_data_chunks(self, f):
        data = f.read()
        # find all 'MDbData' chunks
        regex_pattern = re.compile(b"MDbData")
        matches = regex_pattern.finditer(data)
        positions = []
        for match in matches:
            positions.append(match.start())
        return positions

    def extract_laterality_data(self, f):
        raw = f.read(20)
        try:
            laterality_data = self.lat_structure.parse(raw)
            if laterality_data.laterality == 82:
                laterality = "R"
            elif laterality_data.laterality == 76:
                laterality = "L"
        except:
            laterality = None
        return laterality

    def read_oct_volume(self):
        """ Reads oct data.
            Returns:
                obj:OCTVolumeWithMetaData
        """
        with open(self.filepath, "rb") as f:

            chunk_positions = self.find_data_chunks(f)

            volume_dict = {}
            for start in chunk_positions:
                f.seek(start)
                raw = f.read(60)
                chunk = self.chunk_structure.parse(raw)

                if chunk.type == 11:  # laterality data
                    self.laterality = self.extract_laterality_data(f)

                if chunk.type == 1073741824:  # image data
                    if chunk.ind == 1:  # oct data
                        volume_string = f"{chunk.patient_id}_{chunk.study_id}_{chunk.series_id}"
                        # read data
                        raw = f.read(20)
                        image_data = self.image_structure.parse(raw)
                        all_bits = [
                            f.read(2)
                            for i in range(
                                image_data.height * image_data.width
                            )
                        ]
                        raw_volume = list(
                            map(self.read_custom_float, all_bits)
                        )
                        image = np.array(raw_volume).reshape(
                            image_data.width, image_data.height
                        )
                        normalized_float = pow(image, 1.0 / 2.4) / pow(
                            2, 1.0 / 2.4
                        )
                        image = (normalized_float * (256 * 256 - 1)).astype(
                            np.uint16
                        )
                        if volume_string not in volume_dict:
                            volume_dict[volume_string] = {}
                        volume_dict[volume_string][
                            int(chunk.slice_id / 2)
                        ] = image

            oct_volumes = []
            for key, volume in volume_dict.items():
                slice_order = sorted([id for id in volume.keys()])
                ordered_volume = [volume[id] for id in slice_order]
                oct_volumes.append(
                    OCTVolumeWithMetaData(
                        volume=ordered_volume,
                        patient_id=key,
                        laterality=self.laterality,
                    )
                )

        return oct_volumes

    def read_fundus_image(self):
        """ Reads fundus data.

            Returns:
                obj:FundusImageWithMetaData
        """
        with open(self.filepath, "rb") as f:
            chunk_positions = self.find_data_chunks(f)
            # initalise dict to hold all the image volumes
            image_array_dict = {}
            # traverse all chunks and extract slices
            for start in chunk_positions:
                f.seek(start)
                raw = f.read(60)
                chunk = self.chunk_structure.parse(raw)

                if chunk.type == 11:  # laterality data
                    self.laterality = self.extract_laterality_data(f)

                if chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = self.image_structure.parse(raw)

                    if chunk.ind == 0:  # fundus data
                        raw_volume = np.frombuffer(
                            f.read(image_data.height * image_data.width),
                            dtype=np.uint8,
                        )
                        image = np.array(raw_volume).reshape(
                            image_data.height, image_data.width
                        )
                        image_string = f"{chunk.patient_id}_{chunk.study_id}_{chunk.series_id}"
                        image_array_dict[image_string] = image

            fundus_images = []
            for key, image in image_array_dict.items():
                fundus_images.append(
                    FundusImageWithMetaData(
                        image=image, patient_id=key, laterality=self.laterality
                    )
                )

        return fundus_images

    def read_custom_float(self, bytes):
        """ Implementation of bespoke float type used in .e2e files.

        Notes:
            Custom float is a floating point type with no sign, 6-bit exponent, and 10-bit mantissa.

        Args:
            bytes (str): The two bytes.

        Returns:
            float
        """
        power = pow(2, 10)
        # convert two bytes to 16-bit binary representation
        bits = (
            bin(bytes[0])[2:].zfill(8)[::-1] + bin(bytes[1])[2:].zfill(8)[::-1]
        )

        # get mantissa and exponent
        mantissa = bits[:10]
        exponent = bits[10:]

        # convert to decimal representations
        mantissa_sum = 1 + int(mantissa, 2) / power
        exponent_sum = int(exponent[::-1], 2) - 63
        decimal_value = mantissa_sum * pow(2, exponent_sum)
        return decimal_value



import logging
from pathlib import Path

import click

from panimg import __version__, convert, logger


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__, "-v", "--version")
def cli():
    pass


@cli.command(name="convert", short_help="Convert a directory of image files")
@click.option("-v", "--verbose", count=True)
@click.argument(
    "input",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=False,
        resolve_path=True,
    ),
)
@click.argument(
    "output",
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        resolve_path=True,
    ),
)
def convert_cli(input: str, output: str, verbose: int):
    input_directory = Path(input)
    output_directory = Path(output)

    ch = logging.StreamHandler()

    if verbose == 0:
        logger.setLevel(logging.WARNING)
        ch.setLevel(logging.WARNING)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
        ch.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)

    logger.addHandler(ch)

    if output_directory.exists():
        raise click.exceptions.UsageError(
            f"Output directory {output_directory} already exists"
        )
    else:
        output_directory.mkdir(parents=True)

    convert(input_directory=input_directory, output_directory=output_directory)



import logging
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, Iterable, List, Optional, Set

from panimg.exceptions import UnconsumedFilesException
from panimg.image_builders import DEFAULT_IMAGE_BUILDERS
from panimg.models import (
    PanImg,
    PanImgFile,
    PanImgFolder,
    PanImgResult,
    PostProcessorResult,
)
from panimg.post_processors import DEFAULT_POST_PROCESSORS
from panimg.types import ImageBuilder, PostProcessor

logger = logging.getLogger(__name__)


def convert(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Optional[Iterable[ImageBuilder]] = None,
    post_processors: Optional[Iterable[PostProcessor]] = None,
    recurse_subdirectories: bool = True,
) -> PanImgResult:
    new_images: Set[PanImg] = set()
    new_image_files: Set[PanImgFile] = set()
    new_folders: Set[PanImgFolder] = set()
    consumed_files: Set[Path] = set()
    file_errors: DefaultDict[Path, List[str]] = defaultdict(list)

    builders = builders if builders is not None else DEFAULT_IMAGE_BUILDERS

    builder_names = ", ".join(str(b) for b in builders)
    logger.info(f"Using builders {builder_names}")

    _convert_directory(
        input_directory=input_directory,
        output_directory=output_directory,
        builders=builders,
        consumed_files=consumed_files,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=new_folders,
        file_errors=file_errors,
        recurse_subdirectories=recurse_subdirectories,
    )

    result = _post_process(
        image_files=new_image_files,
        post_processors=post_processors
        if post_processors is not None
        else DEFAULT_POST_PROCESSORS,
    )
    new_image_files |= result.new_image_files
    new_folders |= result.new_folders

    return PanImgResult(
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=new_folders,
        consumed_files=consumed_files,
        file_errors=file_errors,
    )


def _convert_directory(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Iterable[ImageBuilder],
    consumed_files: Set[Path],
    new_images: Set[PanImg],
    new_image_files: Set[PanImgFile],
    new_folders: Set[PanImgFolder],
    file_errors: DefaultDict[Path, List[str]],
    recurse_subdirectories: bool = True,
):
    input_directory = Path(input_directory).resolve()
    output_directory = Path(output_directory).resolve()

    output_directory.mkdir(exist_ok=True)

    files = set()
    for o in Path(input_directory).iterdir():
        if o.is_dir() and recurse_subdirectories:
            _convert_directory(
                input_directory=input_directory
                / o.relative_to(input_directory),
                output_directory=output_directory
                / o.relative_to(input_directory),
                builders=builders,
                consumed_files=consumed_files,
                new_images=new_images,
                new_image_files=new_image_files,
                new_folders=new_folders,
                file_errors=file_errors,
                recurse_subdirectories=recurse_subdirectories,
            )
        elif o.is_file():
            files.add(o)

    logger.info(f"Converting directory {input_directory}")

    for builder in builders:
        builder_files = files - consumed_files

        if len(builder_files) == 0:
            return

        logger.debug(f"Processing {len(builder_files)} file(s) with {builder}")

        builder_result = _build_files(
            builder=builder,
            files=builder_files,
            output_directory=output_directory,
        )

        new_images |= builder_result.new_images
        new_image_files |= builder_result.new_image_files
        new_folders |= builder_result.new_folders
        consumed_files |= builder_result.consumed_files

        if builder_result.consumed_files:
            logger.info(
                f"{builder} created {len(builder_result.new_images)} "
                f"new images(s) from {len(builder_result.consumed_files)} file(s)"
            )
        else:
            logger.debug(f"No files consumed by {builder}")

        for filepath, errors in builder_result.file_errors.items():
            file_errors[filepath].extend(errors)


def _build_files(
    *, builder: ImageBuilder, files: Set[Path], output_directory: Path,
) -> PanImgResult:
    new_images = set()
    new_image_files: Set[PanImgFile] = set()
    consumed_files: Set[Path] = set()
    file_errors: Dict[Path, List[str]] = {}

    try:
        for result in builder(files=files):
            n_image, n_image_files = result.save(
                output_directory=output_directory
            )

            new_images.add(n_image)
            new_image_files |= n_image_files
            consumed_files |= result.consumed_files

    except UnconsumedFilesException as e:
        file_errors = e.file_errors

    return PanImgResult(
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=set(),
        consumed_files=consumed_files,
        file_errors=file_errors,
    )


def _post_process(
    *, image_files: Set[PanImgFile], post_processors: Iterable[PostProcessor]
) -> PostProcessorResult:
    new_image_files: Set[PanImgFile] = set()
    new_folders: Set[PanImgFolder] = set()

    logger.info(f"Post processing {len(image_files)} image(s)")

    for processor in post_processors:
        result = processor(image_files=image_files)
        new_image_files |= result.new_image_files
        new_folders |= result.new_folders

    return PostProcessorResult(
        new_image_files=new_image_files, new_folders=new_folders
    )



from typing import Iterable

from panimg.image_builders.dicom import image_builder_dicom
from panimg.image_builders.fallback import image_builder_fallback
from panimg.image_builders.metaio_mhd_mha import image_builder_mhd
from panimg.image_builders.nifti import image_builder_nifti
from panimg.image_builders.nrrd import image_builder_nrrd
from panimg.image_builders.oct import image_builder_oct
from panimg.image_builders.tiff import image_builder_tiff
from panimg.types import ImageBuilder

DEFAULT_IMAGE_BUILDERS: Iterable[ImageBuilder] = [
    image_builder_mhd,
    image_builder_nifti,
    image_builder_nrrd,
    image_builder_dicom,
    image_builder_tiff,
    image_builder_oct,
    image_builder_fallback,
]

__all__ = [
    "image_builder_mhd",
    "image_builder_nifti",
    "image_builder_nrrd",
    "image_builder_dicom",
    "image_builder_tiff",
    "image_builder_oct",
    "image_builder_fallback",
    "DEFAULT_IMAGE_BUILDERS",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['panimg',
 'panimg.contrib',
 'panimg.contrib.oct_converter',
 'panimg.contrib.oct_converter.image_types',
 'panimg.contrib.oct_converter.readers',
 'panimg.image_builders',
 'panimg.post_processors']

package_data = \
{'': ['*']}

install_requires = \
['Pillow',
 'SimpleITK',
 'click',
 'construct',
 'numpy',
 'openslide-python',
 'pydantic',
 'pydicom>=2.2',
 'pylibjpeg',
 'pylibjpeg-libjpeg',
 'pylibjpeg-openjpeg',
 'pylibjpeg-rle',
 'pyvips',
 'tifffile']

entry_points = \
{'console_scripts': ['panimg = panimg.cli:cli']}

setup_kwargs = {
    'name': 'panimg',
    'version': '0.5.2',
    'description': 'Conversion of medical images to MHA and TIFF.',
    'long_description': '# panimg\n\n[![CI](https://github.com/DIAGNijmegen/rse-panimg/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/DIAGNijmegen/rse-panimg/actions/workflows/ci.yml?query=branch%3Amain)\n[![PyPI](https://img.shields.io/pypi/v/panimg)](https://pypi.org/project/panimg/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/panimg)](https://pypi.org/project/panimg/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**NOT FOR CLINICAL USE**\n\nConversion of medical images to MHA and TIFF. \nRequires Python 3.7, 3.8 or 3.9.\n`libvips-dev` and `libopenslide-dev` must be installed on your system.\n\nUnder the hood we use:\n\n* `SimpleITK`\n* `pydicom`\n* `pylibjpeg`  \n* `Pillow`\n* `openslide-python`\n* `pyvips`\n* `oct-converter`\n\n## Usage\n\n`panimg` takes a folder and tries to convert the containing files to MHA or TIFF.\nBy default, it will try to convert files from subdirectories as well.\nTo only convert files in the top level directory, set `recurse_subdirectories` to `False`.\nIt will try several strategies for loading the contained files, and if an image is found it will output it to the output folder.\nIt will return a structure containing information about what images were produced, what images were used to form the new images, image metadata, and any errors from any of the strategies.\n\n\n**NOTE: Alpha software, do not run this on folders you do not have a backup of.**\n\n```python\nfrom pathlib import Path\nfrom panimg import convert\n\nresult = convert(\n    input_directory=Path("/path/to/files/"),\n    output_directory=Path("/where/files/will/go/"),\n)\n```\n\n### Command Line Interface\n\n`panimg` is also accessible from the command line.\nInstall the package from pip as before, then you can use:\n\n**NOTE: Alpha software, do not run this on folders you do not have a backup of.**\n\n```shell\npanimg convert /path/to/files/ /where/files/will/go/\n```\n\nTo access the help test you can use `panimg -h`.\n\n### Supported Formats\n\n| Input                               | Output  | Strategy   | Notes                      |\n| ----------------------------------- | --------| ---------- | -------------------------- |\n| `.mha`                              | `.mha`  | `metaio`   |                            |\n| `.mhd` with `.raw` or `.zraw`       | `.mha`  | `metaio`   |                            |\n| `.dcm`                              | `.mha`  | `dicom`    |                            |\n| `.nii`                              | `.mha`  | `nifti`    |                            |\n| `.nii.gz`                           | `.mha`  | `nifti`    |                            |\n| `.nrrd`                             | `.mha`  | `nrrd`     | <sup>[1](#footnote1)</sup> |\n| `.e2e`                              | `.mha`  | `oct`      | <sup>[2](#footnote2)</sup> |\n| `.fds`                              | `.mha`  | `oct`      | <sup>[2](#footnote2)</sup> |\n| `.fda`                              | `.mha`  | `oct`      | <sup>[2](#footnote2)</sup> |\n| `.png`                              | `.mha`  | `fallback` | <sup>[3](#footnote3)</sup> |\n| `.jpeg`                             | `.mha`  | `fallback` | <sup>[3](#footnote3)</sup> |\n| `.tiff`                             | `.tiff` | `tiff`     |                            |\n| `.svs` (Aperio)                     | `.tiff` | `tiff`     |                            |\n| `.vms`, `.vmu`, `.ndpi` (Hamamatsu) | `.tiff` | `tiff`     |                            |\n| `.scn` (Leica)                      | `.tiff` | `tiff`     |                            |\n| `.mrxs` (MIRAX)                     | `.tiff` | `tiff`     |                            |\n| `.biff` (Ventana)                   | `.tiff` | `tiff`     |                            |\n\n<a name="footnote1">1</a>: Detached headers are not supported.\n\n<a name="footnote2">2</a>: Only OCT volume(s), no fundus image(s) will be extracted.\n\n<a name="footnote3">3</a>: 2D only, unitary dimensions\n\n#### Post Processors\n\nYou can also define a set of post processors that will operate on each output file.\nWe provide a `dzi_to_tiff` post processor that is enabled by default, which will produce a DZI file if it is able to.\nTo customise the post processors that run you can do this with\n\n```python\nresult = convert(..., post_processors=[...])\n```\n\n#### Using Strategies Directly\n\nIf you want to run a particular strategy directly which returns a generator of images for a set of files you can do this with\n\n```python\nfiles = {f for f in Path("/foo/").glob("*.dcm") if f.is_file()}\n\ntry:\n    for result in image_builder_dicom(files=files):\n        sitk_image = result.image\n        process(sitk_image)  # etc. you can also look at result.name for the name of the file,\n                             # and result.consumed_files to see what files were used for this image\nexcept UnconsumedFilesException as e:\n    # e.errors is keyed with a Path to a file that could not be consumed,\n    # with a list of all the errors found with loading it,\n    # the user can then choose what to do with that information\n    ...\n```\n',
    'author': 'James Meakin',
    'author_email': '12661555+jmsmkn@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DIAGNijmegen/rse-panimg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
